subroutine qumc(&
     Ndata,lam2,q,u,q_err,u_err,&                    ! data
     Niter,&                                         ! MCMC setting
     src_delta,src_gauss,src_boxcar,Npar,par,peri,&  ! fitting model
     chain,Nacpt,Nstep)                              ! output

  implicit none

  double precision, parameter   :: pi  = acos(-1.d0)
  double precision, parameter   :: ipi = 1.d0/pi

  integer(4)      , intent(in)  :: Ndata
  double precision, intent(in)  :: lam2(0:Ndata-1), q(0:Ndata-1), u(0:Ndata-1), q_err(0:Ndata-1), u_err(0:Ndata-1)

  integer(4)      , intent(in)  :: Niter

  integer(4)      , intent(in)  :: src_delta, src_gauss, src_boxcar, Npar
  double precision, intent(in)  :: par(0:3,0:Npar-1) !0:init, 1:dx, 2:min, 3:max
  integer(4)      , intent(in)  :: peri(0:Npar-1)

  double precision, intent(out) :: chain(0:Npar,0:Niter-1)
  integer(4)      , intent(out) :: Nacpt(0:Npar-1), Nstep(0:Npar-1)

  integer(4)                    :: iiter, ipar, i

  integer(4)                    :: flag
  double precision              :: candidate(0:Npar-1)
  double precision              :: dum_chain(0:Npar)
  double precision              :: logu, logr
  double precision              :: chisq, chisq_old

  ! Initialization
  Nacpt = 0
  Nstep = 0

  chain(0:Npar-1,0) = par(0,:)
  chain(Npar,0) = calc_chisq(Ndata,Npar,lam2,q,u,q_err,u_err,src_delta,src_gauss,src_boxcar,chain(0:Npar-1,0))


  ! MCMC
  roop_iter:do iiter = 0, Niter-1

     roop_params:do ipar = 0, Npar-1

        i = rand_int_unif(0,Npar-1)

        candidate(0:Npar-1) = chain(0:Npar-1,iiter)
        chisq_old           = chain(Npar,iiter)
        candidate(i)        = candidate(i) + rand_dble_norm(0.d0,par(1,i))

        ! If the candidate parameter is initial angles of FDF (peri=1),
        ! the parameter moves peiodically with the period [-pi/2,pi/2],
        ! and 'flag' is always set to 1.
        ! Else, check if the candidate parameter is inside a given boundary condition.
        ! flag =  1 :  inside boundary
        ! flag = -1 : outside boundary
        call check_p(peri(i),par(2,i),par(3,i),candidate(i),flag)

        ! if the candidate parameter is outside of a given boundary,
        ! reject it.
        if(flag < 0) then
           Nstep(i) = Nstep(i) + 1

        else
           ! calculate a chisquare value
           chisq = calc_chisq(Ndata,Npar,lam2,q,u,q_err,u_err,src_delta,src_gauss,src_boxcar,candidate)

           ! calculate a Metropolis ratio
           logr = -5.d-1 * (chisq - chisq_old)

           ! accept the candidate parameter
           ! with the probability equal to the Metropolis ratio
           logu = log(rand_num())
           !print *, chisq, chisq_old, logr, logu
           if ( (logu < logr) .or. (logr >= 0) ) then
              Nstep(i)              = Nstep(i) + 1
              Nacpt(i)              = Nacpt(i) + 1
              chain(0:Npar-1,iiter) = candidate(0:Npar-1)
              chain(Npar,iiter)     = chisq
           else
              Nstep(i)              = Nstep(i) + 1
           end if
        end if

     end do roop_params

     if(iiter < Niter-1) then
        chain(0:Npar,iiter+1) = chain(0:Npar,iiter)
     end if

  end do roop_iter



contains

  subroutine check_p(peri,parmin,parmax,candidate,flag)
    implicit none

    integer(4)      , intent(in)    :: peri
    double precision, intent(in)    :: parmin, parmax
    double precision, intent(inout) :: candidate
    integer(4)      , intent(out)   :: flag

    double precision                :: factor, sign

    if(peri.eq.1) then
       flag = 1

       factor    = int(abs(candidate)*ipi+5.d-1)
       sign      = abs(candidate)/candidate
       candidate = candidate - pi*sign*factor

    else
       ! factor = min(candidate-parmin,parmax-candidate)
       ! sign   = abs(factor)/factor
       ! flag   = int(sign)
       if(candidate.gt.parmin .and. candidate.lt.parmax) then
          flag =  1
       else
          flag = -1
       end if
    end if

  end subroutine check_p


  function calc_chisq(Ndata,Npar,lam2,q,u,q_err,u_err,src_delta,src_gauss,src_boxcar,candidate)
    implicit none

    integer(4)      , intent(in) :: Ndata, Npar, src_delta, src_gauss, src_boxcar
    double precision, intent(in) :: lam2(0:Ndata-1), q(0:Ndata-1), u(0:Ndata-1), q_err(0:Ndata-1), u_err(0:Ndata-1)
    double precision, intent(in) :: candidate(0:Npar-1)

    double precision             :: calc_chisq

    integer(4)                   :: i, j, g, b
    double precision             :: delta
    complex(kind(0d0))           :: PI_sum

    g = 3*src_delta
    b = 3*src_delta + 4*src_gauss
    calc_chisq = 0.d0
    do j = 0, Ndata-1
       PI_sum = 0.d0
       if(src_delta.ne.0) then
          do i = 0, src_delta-1
             PI_sum = PI_sum &
                  + P_delta(lam2(j),candidate(3*i),candidate(3*i+1),candidate(3*i+2))
          end do
       end if
       if(src_gauss.ne.0) then
          do i = 0, src_gauss-1
             PI_sum = PI_sum &
                  + P_gauss(lam2(j),candidate(4*i+g),candidate(4*i+1+g),candidate(4*i+2+g),candidate(4*i+3+g))
          end do
       end if
       if(src_boxcar.ne.0) then
          do i = 0, src_boxcar-1
             PI_sum = PI_sum &
                  + P_boxcar(lam2(j),candidate(4*i+b),candidate(4*i+1+b),candidate(4*i+2+b),candidate(4*i+3+b))
          end do
       end if

       delta = (real(PI_sum)-q(j))**2/q_err(j)**2 + (aimag(PI_sum)-u(j))**2/u_err(j)**2

       calc_chisq = calc_chisq + delta

    end do
  end function calc_chisq


  ! lam  : (wavelength [m])^2
  ! phi0 : Faraday depth [rad/m^2]
  ! amp0 : intensity
  ! evpa0: electric vector polarization angle [radian]
  ! dphi0: width in Faraday depth space [rad/m^2]

  ! Delta function
  function P_delta(lambda,phi0,amp0,evpa0)
    implicit none
    double precision , intent(in) :: lambda
    double precision , intent(in) :: phi0, amp0, evpa0
    complex(kind(0d0))            :: P_delta
    P_delta = amp0*cmplx(cos(2.d0*phi0*lambda+2.d0*evpa0), sin(2.d0*phi0*lambda+2.d0*evpa0))
    ! if(abs( real(P_delta)).le.1.d-60) P_delta = cmplx(0.d0         ,aimag(P_delta))
    ! if(abs(aimag(P_delta)).le.1.d-60) P_delta = cmplx(real(P_delta),0.d0          )
  end function P_delta

  ! Gaussian function
  function P_gauss(lambda,phi0,amp0,evpa0,dphi0)
    implicit none
    double precision , intent(in) :: lambda
    double precision , intent(in) :: phi0, amp0, evpa0, dphi0
    complex(kind(0d0))            :: P_gauss
    P_gauss = amp0*exp(-2.d0*dphi0**2*lambda**2) &
         * cmplx(cos(2.d0*phi0*lambda+2.d0*evpa0), sin(2.d0*phi0*lambda+2.d0*evpa0))
    ! if(abs( real(P_gauss)).le.1.d-60) P_gauss = cmplx(0.d0         ,aimag(P_gauss))
    ! if(abs(aimag(P_gauss)).le.1.d-60) P_gauss = cmplx(real(P_gauss),0.d0          )
  end function P_gauss

  ! Boxcar (top-hat) function
  function P_boxcar(lambda,phi0,amp0,evpa0,dphi0)
    implicit none
    double precision , intent(in) :: lambda
    double precision , intent(in) :: phi0, amp0, evpa0, dphi0
    complex(kind(0d0))            :: P_boxcar
    ! P_boxcar = f*sin(dphi*lambda)/(dphi*lambda) &
    !      * cmplx(cos(2.d0*phi*lambda+2.d0*ia), sin(2.d0*phi*lambda+2.d0*ia))
    P_boxcar = amp0*sin(dphi0*lambda)/(lambda) &
         * cmplx(cos(2.d0*phi0*lambda+2.d0*evpa0), sin(2.d0*phi0*lambda+2.d0*evpa0))
    ! if(abs( real(P_boxcar)).le.1.d-60) P_boxcar = cmplx(0.d0         ,aimag(P_boxcar))
    ! if(abs(aimag(P_boxcar)).le.1.d-60) P_boxcar = cmplx(real(P_boxcar),0.d0          )
  end function P_boxcar


  function rand_num()
    implicit none
    double precision             :: rand_num
    call random_number(rand_num)
  end function rand_num


  function rand_int_unif(rmin,rmax)
    implicit none
    integer         , intent(in) :: rmin, rmax
    integer                      :: rand_int_unif
    rand_int_unif = int(dble(rmax-rmin+1)*rand_num()) + rmin
  end function rand_int_unif


  function rand_dble_norm(mu,var)
    implicit none
    double precision, intent(in) :: mu, var
    double precision             :: rand_dble_norm
    double precision             :: rnd1, rnd2
    call random_number(rnd1)
    call random_number(rnd2)
    rand_dble_norm = sqrt(-2.d0*log(rand_num()))*cos(2.d0*pi*rand_num()) * var + mu
  end function rand_dble_norm

end subroutine qumc



subroutine init_rand_seed(myseed)
  implicit none
  integer(4)        , intent(in)  :: myseed
  integer                         :: nRand
  integer           , allocatable :: seed(:)
  call random_seed(size = nRand)
  allocate(seed(nRand))
  seed = myseed
  call random_seed(put = seed)
end subroutine init_rand_seed



subroutine init_rand_clock(myseed)
  implicit none
  integer(4)        , intent(out) :: myseed(0:0)
  integer                         :: clock
  integer                         :: nRand
  integer           , allocatable :: seed(:)
  call random_seed(size = nRand)
  allocate(seed(nRand))
  call system_clock(count = clock)
  seed = clock
  myseed = clock
  call random_seed(put = seed)
end subroutine init_rand_clock

