"""RM module that implements 1-D RM Synthesis/RMCLEAN."""

"""
MIT License

Copyright (c) 2017 George Heald

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# Written by George Heald
# v1.0, 14 November 2017

# Modified by Shinsuke Ideguchi (9 May 2018)

from numpy import *
from pylab import *
from scipy.optimize import curve_fit

class PolObservation:
	"""Class to describe an observation & perform polarimetry operations"""

	def __init__(self,l2,QU,QUerr=None,verbose=True):
		"""
		Initialise a PolObservation object.

		The object needs to know the frequency setup and
		obtain Stokes IQU data.
		If you don't have I, but only QU, just provide ones.
		Optionally, errors can be provided for I, Q, and U.

		Parameters
		----------
		freq : array
		    Frequencies in Hz
		QU : array
		    Stokes QU values in a 2xN array where
		    N is the length of freq.
		QUerr : array, optional (default None)
		    Stokes QU uncertainties in a 2xN array
		verbose : boolean, optional (default True)
		    Print some output?

		#When initialised, a powerlaw will be fit to Stokes I
		#for possible later use.
		"""

                self.l2 = l2
		self.q = QU[0]
		self.u = QU[1]
		if QUerr is None:
			self.qerr = None
			self.uerr = None
		else:
			self.qerr = QUerr[0]
			self.uerr = QUerr[1]
		# self.i_model = p[0]*(freq/mean(freq))**p[1]
		# if verbose: print 'Fitted spectral index is %f'%p[1]
		# self.model_s = p[0]
		# self.model_a = p[1]
		self.rmsynth_done = False
		self.rmclean_done = False

	def rmsynthesis(self,phi,double=True,pclip=None,verbose=True):
		"""
		Perform RM Synthesis.

		This function performs RM Synthesis on the provided
		IQU data and computes the RMSF.
		It can adopt weights based on the IQU errors if they
		are provided.

		Parameters
		----------
		phi : array
		   RM values to use (should be contiguous and regularly spaced)
		double : boolean, optional (default True)
		   Create the RMSF double the length along the RM axis?
		   This should be kept as True unless you are not planning
		   to RMCLEAN.
		pclip : float, optional (default -inf)
		   Ignore channels with sqrt(Q**2+U**2) S/N ratio < pclip
		verbose : boolean, optional (default True)
		    Print some output?

		"""

		# Constants
		c = 2.998e8
		ci = complex(0.,1.)

		# Compute lamba-squared values
		# l2 = (c/self.freq)**2
                print(self.l2)

		# Clip values if needed
		if pclip is None: pclip = -inf

                gp = full(len(self.q),True, dtype=bool)
		if verbose: print 'Using %d of %d channels'%(sum(gp),len(self.l2))

		# Work out weights if needed and overall normalisation
		print 'Using uniform weights'
		weights = ones(self.q.shape)

                ### added
                # dl2 = empty(self.freq.shape,dtype=float)
                # for j,tmpl2 in enumerate(l2):
                #         if j==0:
                #                 dl2[0] = abs(l2[1]-l2[0])
                #         elif j==len(l2)-1:
                #                 dl2[len(l2)-1] = abs(l2[-1]-l2[-2])
                #         else:
                #                 dl2[j] = abs(l2[j+1]-l2[j-1])*.5

		# K = 1./sum(weights[gp]*dl2[gp]) # Normalisation
                ###
		K = 1./sum(weights[gp]) # Normalisation
		if verbose: print 'Normalisation value = %f'%(K)

		# Calculate and report FWHM and other metrics
		min_l2 = min(self.l2[gp])
		min_dl2 = min(abs(diff(self.l2[gp])))
		fwhm = 2.*sqrt(3.)/(max(self.l2[gp])-min(self.l2[gp]))
		maxscale = pi/min_l2
		rm_max = sqrt(3.)/min_dl2
		if verbose:
			print 'FWHM of RMSF is %f rad/m2'%fwhm
			print 'Max RM scale is %f rad/m2'%maxscale
			print 'Max RM value is %f rad/m2'%rm_max
		l20 = mean(self.l2[gp])

		# QU values to operate on, normalized if needed
		quvec = self.q + ci*self.u

		# Work out dimensions of RMSF
		dphi = min(diff(phi))
		if double:
			nrmsf = len(phi)*2 + 1
		else:
			nrmsf = len(phi)
		rmsf_phi = linspace(-float(nrmsf/2)*dphi,float(nrmsf/2)*dphi,nrmsf)

		rmsf = zeros(rmsf_phi.shape,dtype=complex)
		fdf = zeros(phi.shape,dtype=complex)

		# Now do the work
		if verbose: print 'Calculating RMSF...'
		for j,phival in enumerate(rmsf_phi):
			rmsf[j]=K*sum(weights[gp]*exp(-2.*ci*phival*(self.l2[gp]-l20)))
			### rmsf[j]=K*sum(weights[gp]*exp(-2.*ci*phival*(l2[gp]-l20))*dl2[gp])
		if verbose: print 'Calculating FDF...'
		for j,phival in enumerate(phi):
			fdf[j]=K*sum(quvec[gp]*weights[gp]*exp(-2.*ci*phival*(self.l2[gp]-l20)))
			### fdf[j]=K*sum(quvec[gp]*weights[gp]*exp(-2.*ci*phival*(l2[gp]-l20))*dl2[gp])

		# For later use
		self.l20 = l20
		self.weights = weights
		self.K = K
		self.rmsf = rmsf
		self.rmsf_phi = rmsf_phi
		self.fdf = fdf
		self.rmsf_fwhm = fwhm
		self.maxscale = maxscale
		self.rm_max = rm_max
		self.dphi = dphi
		self.rmsf_fwhm_pix = int(fwhm/dphi+0.5)
		self.phi = phi
		self.nphi = len(phi)
		self.rmsynth_done = True
		self.rmclean_done = False

	def plot_fdf(self,display=True,save=None):
		"""
		Plot the FDF and RMSF

		The plot will show the RMSF in black, and the FDF in red.
		If RMCLEAN has already been performed then the cleaned spectrum
		will be shown in blue, and the clean model in green.

		Parameters
		----------
		display : boolean, optional (default True)
		       Show plot on screen?
		save : str, optional (default None)
		       Save figure to disk? Provide filename if desired.

		"""

		if self.rmsynth_done:
                        # f = loadtxt('data/galaxy_fdf.txt')
                        # phi = f[:,0]
                        # fdf = f[:,1] + f[:,2]*1j

			figure()
			#plot(self.rmsf_phi,abs(self.rmsf),'k-')
			#plot(phi,abs(fdf)/np.pi,'k-')
			plot(self.phi,abs(self.fdf),'r-')
			if self.rmclean_done:
				plot(self.phi,abs(self.rm_cleaned),'b-')
				plot(self.phi,abs(self.rm_comps),'g-')
				legend(('FDF','Clean','Comps'),loc='best')
			else:
				legend('FDF',loc='best')
			xlabel('RM (rad/m2)')
			ylabel('Amplitude')
			if save is not None: savefig(save,bbox_inches='tight')
			if display: show()
		#close('all')

	def plot_clean_proc(self):
		"""
		Generate gif file displaying CLEAN iteration

		"""

                fig = figure()
                my_ylim = max(max(abs(self.my_rm_cleaned[-1,:])),max(abs(self.fdf)))
                for i in range(len(self.my_rm_resid)):
			plot(self.phi,abs(self.fdf),'r-')
                        plot(self.phi,self.my_rm_resid[i,:],'grey')
                        plot(self.phi,abs(self.my_rm_cleaned[i,:]),'b-')
                        #plot(self.phi,abs(self.my_rm_comps[i,:]),'g-')
                	legend(('FDF','Residual','Clean'),loc='best')
                        title('Iteration: '+str(i+1)+'/'+str(len(self.my_rm_resid)))
                        xlabel('RM (rad/m2)')
		        ylabel('Amplitude')
                        ylim(0,my_ylim)
                        pause(.01)
                        inp = raw_input("\r Enter to continue/'q' to quit: ")
                        if inp == 'q': break
                        close()


	# def plot_stokesi(self,display=True,save=None):
	# 	"""
	# 	Plot Stokes I
	
	# 	The plot will show the Stokes I values (and errors if available).

	# 	Parameters
	# 	----------
	# 	display : boolean, optional (default True)
	# 	       Show plot on screen?
	# 	save : str, optional (default None)
	# 	       Save figure to disk? Provide filename if desired.

	# 	"""

	# 	figure()
	# 	if self.ierr==None:
	# 		plot(self.freq,self.i,marker='o',ls='none')
	# 	else:
	# 		errorbar(self.freq,self.i,yerr=self.ierr,marker='o',ls='none')
	# 	plot(self.freq,self.i_model,'k--')
	# 	xlabel('Frequency (Hz)')
	# 	ylabel('Stokes I')
	# 	if save is not None: savefig(save,bbox_inches='tight')
	# 	if display: show()
	# 	#close('all')

	def get_fdf_peak(self,verbose=True):
		"""
		Obtain the peak of the FDF and its Faraday depth

		This will fit the peak of the FSF and report results
		if requested.
		Values that are calculated and reported are:
			Absolute value at peak, PA at peak, peak RM value
		This is done for the dirty FDF.
		If RMCLEAN has been performed then this is also done for the 
		deconvolved Faraday spectrum.
		
		Parameters
		----------
		verbose : boolean, optional (default True)
		    Print some output?

		"""
		if self.rmsynth_done:
			x0 = where(abs(self.fdf)==max(abs(self.fdf)))[0]
			if x0 == 0 or x0 == len(self.phi)-1:
				self.fdf_peak_rm = self.phi[x0]
				self.fdf_peak = max(abs(self.fdf))
				self.fdf_peak_err = 0.
			else:
				y = abs(self.fdf[x0-1:x0+2])
				x = self.phi[x0-1:x0+2]
				p = polyfit(x,y,2)
				self.fdf_peak_rm = -p[1]/(2.*p[0])
				self.fdf_peak = polyval(p,self.fdf_peak_rm)
				self.fdf_peak_rm_err = self.rmsf_fwhm/(2.355*self.fdf_peak/std(real(self.fdf)))
				rotpeak = self.fdf[x0]*exp(-2.*complex(0.,1.)*self.fdf_peak_rm*self.l20)
				pa = 0.5*arctan2(imag(rotpeak),real(rotpeak))
				self.pa = (pa*180./pi)%360.
			if verbose: print 'FDF peaks at an amplitude %f, at RM=%f +/- %f rad/m2'%(self.fdf_peak,self.fdf_peak_rm,self.fdf_peak_rm_err)
			if verbose: print 'RM-corrected PA of FDF peak is %f degrees'%(self.pa)
		if self.rmclean_done:
			x0 = where(abs(self.rm_cleaned)==max(abs(self.rm_cleaned)))[0]
			if x0 == 0 or x0 == len(self.phi)-1:
				self.cln_fdf_peak_rm = self.phi[x0]
				self.cln_fdf_peak = max(abs(self.fdf))
			else:
				y = abs(self.rm_cleaned[x0-1:x0+2])
				x = self.phi[x0-1:x0+2]
				p = polyfit(x,y,2)
				self.cln_fdf_peak_rm = -p[1]/(2.*p[0])
				self.cln_fdf_peak = polyval(p,self.fdf_peak_rm)
				self.cln_fdf_peak_rm_err = self.rmsf_fwhm/(2.355*self.fdf_peak/std(real(self.rm_resid)))
				cln_rotpeak = self.rm_cleaned[x0]*exp(-2.*complex(0.,1.)*self.fdf_peak_rm*self.l20)
				cln_pa = 0.5*arctan2(imag(cln_rotpeak),real(cln_rotpeak))
				self.cln_pa = (cln_pa*180./pi)%360.
			if verbose: print 'Cleaned FDF peaks at an amplitude %f, at RM=%f +/- %f rad/m2'%(self.cln_fdf_peak,self.cln_fdf_peak_rm,self.cln_fdf_peak_rm_err)
			if verbose: print 'RM-corrected PA of cleaned FDF peak is %f degrees'%(self.cln_pa)
			
	def rmclean(self,niter=1000,gain=0.1,cutoff=2.,mask=False,verbose=True):
		"""
		Perform RMCLEAN

		The FDF will be deconvolved using the RMSF.

		Parameters
		----------
		niter : int, optional (default 1000)
		     Maximum number of clean iterations
		gain : float, optional (default 0.1)
		     Clean gain
		cutoff : float, optional (default 2)
		     Clean cutoff, in units of S/N
		     The default stops at 2*sigma above the mean
		mask : boolean (default False)
		     If True, all clean components must be within an
		     RMSF FWHM of the first peak
		verbose : boolean, optional (default True)
		    Print some output?

		"""

		noise = std(real(self.fdf))
		zerolev = median(abs(self.fdf))
		cleanlim = cutoff*noise+zerolev
		if verbose: print 'CLEAN will proceed down to %f'%(cleanlim)
		num = 0
		res = self.fdf.copy()
		tmpcomp = zeros(res.shape,dtype=complex)
		modcomp = zeros(res.shape,dtype=complex)
		resp = abs(res)
		mr = range(len(resp))
                my_resp = zeros([0,len(res)],dtype=float64)
                my_rm_cleaned0 = zeros([0,len(res)],dtype=float64)
		if 10*self.rmsf_fwhm_pix > len(self.phi):
			kernel = exp(-(self.phi-mean(self.phi))**2/(2.*(self.rmsf_fwhm/2.355)**2))
		else:
			kernel = exp(-arange(-self.rmsf_fwhm*5.,self.rmsf_fwhm*5.,self.dphi)**2/(2.*(self.rmsf_fwhm/2.355)**2))
		while max(resp[mr]) > cleanlim and num < niter:
			maxloc = where(resp[mr]==max(resp[mr]))[0]+mr[0]
			if num==0 and verbose:
				print 'First component found at %f rad/m2'%(self.phi[maxloc])
			if num==0 and mask:
				mr=range(maxloc-self.rmsf_fwhm_pix/2,maxloc+self.rmsf_fwhm_pix/2+1)
				if verbose:
					print 'Masking: Clean components must fall within mask of %d/%d pixels'%(len(mr),len(resp))
					print '(i.e. within RM range %f - %f rad/m2)'%(self.phi[mr[0]],self.phi[mr[-1]])
			num += 1
			if num % 10**int(log10(num)) == 0 and verbose:
				print 'Iteration %d: max residual = %f'%(num,max(resp))
			srmsf = roll(self.rmsf,maxloc-self.nphi)
                        #tmpcomp[maxloc] = res[maxloc]*gain
			modcomp[maxloc] += res[maxloc]*gain
                        tmp_cleaned0 = convolve(modcomp,kernel,mode='same')
			subtr = res[maxloc]*gain*srmsf[:self.nphi]
			res -= subtr
			resp = abs(res)
                        my_resp = append(my_resp,array([resp]),axis=0)
                        my_rm_cleaned0 = append(my_rm_cleaned0,array([tmp_cleaned0]),axis=0)
		if verbose: print 'Convolving clean components...'
		self.rm_model = convolve(modcomp,kernel,mode='same')
		if verbose: print 'Restoring convolved clean components...'
		self.rm_cleaned = self.rm_model + res
		self.rm_comps = modcomp
		self.niters = num
		self.rm_resid = res
		self.rmclean_done = True
		self.my_rm_resid = my_resp
		self.my_rm_cleaned = my_rm_cleaned0

	def print_rmstats(self):
		"""
		Print some stats about the results
		
		Some basic statistics will be reported to the terminal:
			mean of RM clean components,
		        dispersion of RM clean components

		"""

		if self.rmclean_done:
			cp = where(abs(self.rm_comps)>0.)
			#cabs = abs(self.rm_comps[cp])
			#crm = self.phi[cp]
			cabs = abs(self.rm_comps)
			crm = self.phi.copy()
			mcrm = sum(cabs*crm)/sum(cabs)
			crmdisp = sqrt(sum(cabs*(crm-mcrm)**2)/sum(cabs))
			print 'Weighted mean RM of clean components is %f'%mcrm
			print 'Weighted RM dispersion of clean components is %f'%crmdisp

